import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.tensorboard import SummaryWriter

from neuralnets.networks.blocks import UNetConvBlock2D, UNetUpSamplingBlock2D, UNetConvBlock3D, UNetUpSamplingBlock3D
from neuralnets.util.io import print_frm
from neuralnets.util.metrics import jaccard, accuracy_metrics
from neuralnets.util.tools import module_to_device, tensor_to_device, log_scalars, log_images_2d, log_images_3d, \
    augment_samples, get_labels, get_unlabeled, clean_labels


class UNetEncoder2D(nn.Module):
    """
    2D U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu'):
        super(UNetEncoder2D, self).__init__()

        self.features = nn.Sequential()
        self.in_channels = in_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.norm = norm
        self.dropout = dropout
        self.activation = activation

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # convolutional block
            conv_block = UNetConvBlock2D(in_features, out_features, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

            # pooling
            pool = nn.MaxPool2d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = UNetConvBlock2D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, norm=norm,
                                           dropout=dropout, activation=activation)

    def forward(self, inputs):

        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class UNetDecoder2D(nn.Module):
    """
    2D U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, norm='instance', dropout=0.0,
                 activation='relu'):
        super(UNetDecoder2D, self).__init__()

        self.features = nn.Sequential()
        self.out_channels = out_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.skip_connections = skip_connections
        self.norm = norm
        self.dropout = dropout
        self.activation = activation

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections:
                conv_block = UNetConvBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = UNetConvBlock2D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv2d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        encoder_outputs.reverse()

        outputs = inputs
        for i in range(self.levels):
            if self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[i],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class UNet2D(nn.Module):
    """
    2D U-Net

    :param optional in_channels: number of input channels, should be odd
    :param optional coi: indices that correspond to the classes of interest
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout_enc: encoder dropout factor
    :param optional dropout_dec: decoder dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, coi=(0, 1), feature_maps=64, levels=4, skip_connections=True, norm='instance',
                 activation='relu', dropout_enc=0.0, dropout_dec=0.0):
        super(UNet2D, self).__init__()

        self.in_channels = in_channels
        self.c = in_channels // 2
        self.coi = coi
        self.out_channels = len(coi)
        self.feature_maps = feature_maps
        self.levels = levels
        self.norm = norm
        self.encoder_outputs = None
        self.decoder_outputs = None

        # contractive path
        self.encoder = UNetEncoder2D(in_channels, feature_maps=feature_maps, levels=levels, norm=norm,
                                     dropout=dropout_enc, activation=activation)
        # expansive path
        self.decoder = UNetDecoder2D(self.out_channels, feature_maps=feature_maps, levels=levels,
                                     skip_connections=skip_connections, norm=norm, dropout=dropout_dec,
                                     activation=activation)

    def forward(self, inputs):

        # contractive path
        self.encoder_outputs, final_output = self.encoder(inputs)

        # expansive path
        self.decoder_outputs, outputs = self.decoder(final_output, self.encoder_outputs)

        return outputs

    def train_epoch(self, loader, loss_fn, optimizer, epoch, augmenter=None, print_stats=1, writer=None,
                    write_images=False, device=0):
        """
        Trains the network for one epoch
        :param loader: dataloader
        :param loss_fn: loss function
        :param optimizer: optimizer for the loss function
        :param epoch: current epoch
        :param augmenter: data augmenter
        :param print_stats: frequency of printing statistics
        :param writer: summary writer
        :param write_images: frequency of writing images
        :param device: GPU device where the computations should occur
        :return: average training loss over the epoch
        """
        # perform training on GPU/CPU
        module_to_device(self, device)
        self.train()

        # keep track of the average loss during the epoch
        loss_cum = 0.0
        cnt = 0

        # start epoch
        for i, data in enumerate(loader):

            # transfer to suitable device and get labels
            data = tensor_to_device(data, device)
            y = get_labels(data[1], coi=self.coi, dtype=float)

            # filter out unlabeled pixels and include them in augmentation
            y_ = get_unlabeled(data[1], dtype=float)
            data.append(y)
            data.append(y_)

            # perform augmentation and transform to appropriate type
            x, _, y, y_ = augment_samples(data, augmenter=augmenter)
            rep = 0
            rep_max = 10
            while (1 - y_[:, self.c:self.c + 1, ...]).sum() == 0 and rep < rep_max:  # make sure labels are not lost in augmentation, otherwise augment new sample
                x, _, y, y_ = augment_samples(data, augmenter=augmenter)
                rep += 1
                if rep == rep_max:
                    x, _, y, y_ = data
            x = x.float()
            y = y[:, self.c:self.c + 1, ...].round().long()
            # clean labels if necessary (due to augmentations)
            if len(self.coi) > 2:
                y = clean_labels(y, len(self.coi))
            y_ = y_[:, self.c:self.c + 1, ...].bool()

            # zero the gradient buffers
            self.zero_grad()

            # forward prop
            y_pred = self(x)

            # compute loss
            loss = loss_fn(y_pred, y[:, 0, ...], mask=~y_)
            loss_cum += loss.data.cpu().numpy()
            cnt += 1

            # backward prop
            loss.backward()

            # apply one step in the optimization
            optimizer.step()

            # print statistics if necessary
            if i % print_stats == 0:
                print_frm('Epoch %5d - Iteration %5d/%5d - Loss: %.6f' % (
                epoch, i, len(loader.dataset) / loader.batch_size, loss))

        # don't forget to compute the average and print it
        loss_avg = loss_cum / cnt
        print_frm('Epoch %5d - Average train loss: %.6f' % (epoch, loss_avg))

        # log everything
        if writer is not None:

            # always log scalars
            log_scalars([loss_avg], ['train/' + s for s in ['loss-seg']], writer, epoch=epoch)

            # log images if necessary
            if write_images:
                log_images_2d([x[:, self.c:self.c + 1, ...]], ['train/' + s for s in ['x']], writer, epoch=epoch)
                y_pred = F.softmax(y_pred, dim=1)
                for i, c in enumerate(self.coi):
                    if not i == 0:  # skip background class
                        y_p = y_pred[:, i:i + 1, ...].data
                        y_t = (y == i).long()
                        log_images_2d([y_t, y_p],
                                      ['train/' + s for s in ['y_class_%d)' % (c), 'y_pred_class_%d)' % (c)]], writer,
                                      epoch=epoch)

        return loss_avg

    def test_epoch(self, loader, loss_fn, epoch, writer=None, write_images=False, device=0):
        """
        Tests the network for one epoch
        :param loader: dataloader
        :param loss_fn: loss function
        :param epoch: current epoch
        :param writer: summary writer
        :param write_images: frequency of writing images
        :param device: GPU device where the computations should occur
        :return: average testing loss over the epoch
        """
        # perform training on GPU/CPU
        module_to_device(self, device)
        self.eval()

        # keep track of the average loss and metrics during the epoch
        loss_cum = 0.0
        cnt = 0

        # test loss
        y_preds = []
        ys = []
        ys_ = []
        for i, data in enumerate(loader):

            # get the inputs and transfer to suitable device
            x, y = tensor_to_device(data, device)
            y_ = get_unlabeled(y)
            x = x.float()
            y = get_labels(y, coi=self.coi, dtype=int)[:, self.c:self.c + 1, ...]
            y_ = get_labels(y_, coi=[0, 255], dtype=bool)[:, self.c:self.c + 1, ...]

            # forward prop
            y_pred = self(x)

            # compute loss
            loss = loss_fn(y_pred, y[:, 0, ...], mask=~y_)
            loss_cum += loss.data.cpu().numpy()
            cnt += 1

            for b in range(y_pred.size(0)):
                y_preds.append(F.softmax(y_pred, dim=1)[b, ...].view(y_pred.size(1), -1).data.cpu().numpy())
                ys.append(y[b, 0, ...].flatten().cpu().numpy())
                ys_.append(y_[b, 0, ...].flatten().cpu().numpy())

        # prep for metric computation
        y_preds = np.concatenate(y_preds, axis=1)
        ys = np.concatenate(ys)
        ys_ = np.concatenate(ys_)
        w = (1 - ys_).astype(bool)
        js = [jaccard((ys == i).astype(int), y_preds[i, :], w=w) for i in range(1, len(self.coi))]
        ams = [accuracy_metrics((ys == i).astype(int), y_preds[i, :], w=w) for i in range(1, len(self.coi))]

        # don't forget to compute the average and print it
        loss_avg = loss_cum / cnt
        print_frm('Epoch %5d - Average test loss: %.6f' % (epoch, loss_avg))

        # log everything
        if writer is not None:

            # always log scalars
            log_scalars([loss_avg], ['test/' + s for s in ['loss-seg']], writer, epoch=epoch)
            for i, c in enumerate(self.coi):
                if not i == 0:  # skip background class
                    log_scalars([js[i - 1], *(ams[i - 1])], ['test/' + s for s in
                                                             ['jaccard_class_%d)' % (c), 'accuracy_class_%d)' % (c),
                                                              'balanced-accuracy_class_%d)' % (c),
                                                              'precision_class_%d)' % (c), 'recall_class_%d)' % (c),
                                                              'f-score_class_%d)' % (c)]], writer, epoch=epoch)

            # log images if necessary
            if write_images:
                log_images_2d([x[:, self.c:self.c + 1, ...]], ['test/' + s for s in ['x']], writer, epoch=epoch)
                y_pred = F.softmax(y_pred, dim=1)
                for i, c in enumerate(self.coi):
                    if not i == 0:  # skip background class
                        y_p = y_pred[:, i:i + 1, ...].data
                        y_t = (y == i).long()
                        log_images_2d([y_t, y_p],
                                      ['test/' + s for s in ['y_class_%d)' % (c), 'y_pred_class_%d)' % (c)]], writer,
                                      epoch=epoch)

        return np.mean(js)

    def train_net(self, train_loader, test_loader, loss_fn, optimizer, epochs, scheduler=None, test_freq=1,
                  augmenter=None, print_stats=1, log_dir=None, write_images_freq=1, device=0):
        """
        Trains the network
        :param train_loader: data loader with training data
        :param test_loader: data loader with testing data
        :param loss_fn: loss function
        :param optimizer: optimizer for the loss function
        :param epochs: number of training epochs
        :param scheduler: optional scheduler for learning rate tuning
        :param test_freq: frequency of testing
        :param augmenter: data augmenter
        :param print_stats: frequency of logging statistics
        :param log_dir: logging directory
        :param write_images_freq: frequency of writing images
        :param device: GPU device where the computations should occur
        """
        # log everything if necessary
        if log_dir is not None:
            writer = SummaryWriter(log_dir=log_dir)
        else:
            writer = None

        j_max = 0
        for epoch in range(epochs):

            print_frm('Epoch %5d/%5d' % (epoch, epochs))

            # train the model for one epoch
            self.train_epoch(loader=train_loader, loss_fn=loss_fn, optimizer=optimizer, epoch=epoch,
                             augmenter=augmenter, print_stats=print_stats, writer=writer,
                             write_images=epoch % write_images_freq == 0, device=device)

            # adjust learning rate if necessary
            if scheduler is not None:
                scheduler.step(epoch=epoch)

                # and keep track of the learning rate
                writer.add_scalar('learning_rate', float(scheduler.get_lr()[0]), epoch)

            # test the model for one epoch is necessary
            if epoch % test_freq == 0:
                j = self.test_epoch(loader=test_loader, loss_fn=loss_fn, epoch=epoch, writer=writer,
                                    write_images=True, device=device)

                # and save model if higher segmentation performance was obtained
                if j > j_max:
                    j_max = j
                    torch.save(self, os.path.join(log_dir, 'best_checkpoint.pytorch'))

            # save model every epoch
            torch.save(self, os.path.join(log_dir, 'checkpoint.pytorch'))

        writer.close()

    def set_coi(self, coi):
        """
        Adjust the classes of interest of a U-Net, this allows for flexible retraining between classes

        :param coi: tuple indicating the class indices
        """
        # adjust fields
        self.coi = coi
        self.out_channels = len(coi)

        # adjust output channels
        self.decoder.output = nn.Conv2d(self.feature_maps, self.out_channels, kernel_size=1)


class UNetEncoder3D(nn.Module):
    """
    3D U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu'):
        super(UNetEncoder3D, self).__init__()

        self.features = nn.Sequential()
        self.in_channels = in_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.norm = norm
        self.dropout = dropout
        self.activation = activation

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # convolutional block
            conv_block = UNetConvBlock3D(in_features, out_features, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

            # pooling
            pool = nn.MaxPool3d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = UNetConvBlock3D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, norm=norm,
                                           dropout=dropout, activation=activation)

    def forward(self, inputs):

        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class UNetDecoder3D(nn.Module):
    """
    3D U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, norm='instance', dropout=0.0,
                 activation='relu'):
        super(UNetDecoder3D, self).__init__()

        self.features = nn.Sequential()
        self.out_channels = out_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.skip_connections = skip_connections
        self.norm = norm
        self.dropout = dropout
        self.activation = activation

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections:
                conv_block = UNetConvBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = UNetConvBlock3D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv3d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        encoder_outputs.reverse()

        outputs = inputs
        for i in range(self.levels):
            if self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[i],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class UNet3D(nn.Module):
    """
    3D U-Net

    :param optional in_channels: number of input channels
    :param optional coi: indices that correspond to the classes of interest
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout_enc: encoder dropout factor
    :param optional dropout_dec: decoder dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, coi=(0, 1), feature_maps=64, levels=4, skip_connections=True, norm='instance',
                 activation='relu', dropout_enc=0.0, dropout_dec=0.0):
        super(UNet3D, self).__init__()

        self.in_channels = in_channels
        self.coi = coi
        self.out_channels = len(coi)
        self.feature_maps = feature_maps
        self.levels = levels
        self.norm = norm
        self.encoder_outputs = None
        self.decoder_outputs = None

        # contractive path
        self.encoder = UNetEncoder3D(in_channels, feature_maps=feature_maps, levels=levels, norm=norm,
                                     dropout=dropout_enc, activation=activation)
        # expansive path
        self.decoder = UNetDecoder3D(self.out_channels, feature_maps=feature_maps, levels=levels,
                                     skip_connections=skip_connections, norm=norm, dropout=dropout_dec,
                                     activation=activation)

    def forward(self, inputs):

        # contractive path
        self.encoder_outputs, final_output = self.encoder(inputs)

        # expansive path
        self.decoder_outputs, outputs = self.decoder(final_output, self.encoder_outputs)

        return outputs

    def train_epoch(self, loader, loss_fn, optimizer, epoch, augmenter=None, print_stats=1, writer=None,
                    write_images=False, device=0):
        """
        Trains the network for one epoch
        :param loader: dataloader
        :param loss_fn: loss function
        :param optimizer: optimizer for the loss function
        :param epoch: current epoch
        :param augmenter: data augmenter
        :param print_stats: frequency of printing statistics
        :param writer: summary writer
        :param write_images: frequency of writing images
        :param device: GPU device where the computations should occur
        :return: average training loss over the epoch
        """
        # perform training on GPU/CPU
        module_to_device(self, device)
        self.train()

        # keep track of the average loss during the epoch
        loss_cum = 0.0
        cnt = 0

        # start epoch
        for i, data in enumerate(loader):

            # transfer to suitable device and get labels
            data = tensor_to_device(data, device)
            y = get_labels(data[1], coi=self.coi, dtype=float)

            # filter out unlabeled pixels and include them in augmentation
            y_ = get_unlabeled(data[1], dtype=float)
            data.append(y)
            data.append(y_)

            # perform augmentation and transform to appropriate type
            x, _, y, y_ = augment_samples(data, augmenter=augmenter)
            rep = 0
            rep_max = 10
            while (1 - y_).sum() == 0 and rep < rep_max:  # make sure labels are not lost in augmentation, otherwise augment new sample
                x, _, y, y_ = augment_samples(data, augmenter=augmenter)
                rep += 1
                if rep == rep_max:
                    x, _, y, y_ = data
            x = x.float()
            y = y.round().long()
            # clean labels if necessary (due to augmentations)
            if len(self.coi) > 2:
                y = clean_labels(y, len(self.coi))
            y_ = y_.bool()

            # zero the gradient buffers
            self.zero_grad()

            # forward prop
            y_pred = self(x)

            # compute loss
            loss = loss_fn(y_pred, y[:, 0, ...], mask=~y_)
            loss_cum += loss.data.cpu().numpy()
            cnt += 1

            # backward prop
            loss.backward()

            # apply one step in the optimization
            optimizer.step()

            # print statistics if necessary
            if i % print_stats == 0:
                print_frm('Epoch %5d - Iteration %5d/%5d - Loss: %.6f' % (
                epoch, i, len(loader.dataset) / loader.batch_size, loss))

        # don't forget to compute the average and print it
        loss_avg = loss_cum / cnt
        print_frm('Epoch %5d - Average train loss: %.6f' % (epoch, loss_avg))

        # log everything
        if writer is not None:

            # always log scalars
            log_scalars([loss_avg], ['train/' + s for s in ['loss-seg']], writer, epoch=epoch)

            # log images if necessary
            if write_images:
                log_images_3d([x], ['train/' + s for s in ['x']], writer, epoch=epoch)
                y_pred = F.softmax(y_pred, dim=1)
                for i, c in enumerate(self.coi):
                    if not i == 0:  # skip background class
                        y_p = y_pred[:, i:i + 1, ...].data
                        y_t = (y == i).long()
                        log_images_3d([y_t, y_p],
                                      ['train/' + s for s in ['y_class_%d)' % (c), 'y_pred_class_%d)' % (c)]], writer,
                                      epoch=epoch)

        return loss_avg

    def test_epoch(self, loader, loss_fn, epoch, writer=None, write_images=False, device=0):
        """
        Tests the network for one epoch
        :param loader: dataloader
        :param loss_fn: loss function
        :param epoch: current epoch
        :param writer: summary writer
        :param write_images: frequency of writing images
        :param device: GPU device where the computations should occur
        :return: average testing loss over the epoch
        """
        # perform training on GPU/CPU
        module_to_device(self, device)
        self.eval()

        # keep track of the average loss and metrics during the epoch
        loss_cum = 0.0
        cnt = 0

        # test loss
        y_preds = []
        ys = []
        ys_ = []
        for i, data in enumerate(loader):

            # get the inputs and transfer to suitable device
            x, y = tensor_to_device(data, device)
            y_ = get_unlabeled(y)
            x = x.float()
            y = get_labels(y, coi=self.coi, dtype=int)
            y_ = get_labels(y_, coi=[0, 255], dtype=bool)

            # forward prop
            y_pred = self(x)

            # compute loss
            loss = loss_fn(y_pred, y[:, 0, ...], mask=~y_)
            loss_cum += loss.data.cpu().numpy()
            cnt += 1

            for b in range(y_pred.size(0)):
                y_preds.append(F.softmax(y_pred, dim=1)[b, ...].view(y_pred.size(1), -1).data.cpu().numpy())
                ys.append(y[b, 0, ...].flatten().cpu().numpy())
                ys_.append(y_[b, 0, ...].flatten().cpu().numpy())

        # prep for metric computation
        y_preds = np.concatenate(y_preds, axis=1)
        ys = np.concatenate(ys)
        ys_ = np.concatenate(ys_)
        w = (1 - ys_).astype(bool)
        js = [jaccard((ys == i).astype(int), y_preds[i, :], w=w) for i in range(1, len(self.coi))]
        ams = [accuracy_metrics((ys == i).astype(int), y_preds[i, :], w=w) for i in range(1, len(self.coi))]

        # don't forget to compute the average and print it
        loss_avg = loss_cum / cnt
        print_frm('Epoch %5d - Average test loss: %.6f' % (epoch, loss_avg))

        # log everything
        if writer is not None:

            # always log scalars
            log_scalars([loss_avg], ['test/' + s for s in ['loss-seg']], writer, epoch=epoch)
            for i, c in enumerate(self.coi):
                if not i == 0:  # skip background class
                    log_scalars([js[i - 1], *(ams[i - 1])], ['test/' + s for s in
                                                             ['jaccard_class_%d)' % (c), 'accuracy_class_%d)' % (c),
                                                              'balanced-accuracy_class_%d)' % (c),
                                                              'precision_class_%d)' % (c), 'recall_class_%d)' % (c),
                                                              'f-score_class_%d)' % (c)]], writer, epoch=epoch)

            # log images if necessary
            if write_images:
                log_images_3d([x], ['test/' + s for s in ['x']], writer, epoch=epoch)
                y_pred = F.softmax(y_pred, dim=1)
                for i, c in enumerate(self.coi):
                    if not i == 0:  # skip background class
                        y_p = y_pred[:, i:i + 1, ...].data
                        y_t = (y == i).long()
                        log_images_3d([y_t, y_p],
                                      ['test/' + s for s in ['y_class_%d)' % (c), 'y_pred_class_%d)' % (c)]], writer,
                                      epoch=epoch)

        return np.mean(js)

    def train_net(self, train_loader, test_loader, loss_fn, optimizer, epochs, scheduler=None, test_freq=1,
                  augmenter=None, print_stats=1, log_dir=None, write_images_freq=1, device=0):
        """
        Trains the network
        :param train_loader: data loader with training data
        :param test_loader: data loader with testing data
        :param loss_fn: loss function
        :param optimizer: optimizer for the loss function
        :param epochs: number of training epochs
        :param scheduler: optional scheduler for learning rate tuning
        :param test_freq: frequency of testing
        :param augmenter: data augmenter
        :param print_stats: frequency of logging statistics
        :param log_dir: logging directory
        :param write_images_freq: frequency of writing images
        :param device: GPU device where the computations should occur
        """
        # log everything if necessary
        if log_dir is not None:
            writer = SummaryWriter(log_dir=log_dir)
        else:
            writer = None

        j_max = 0
        for epoch in range(epochs):

            print_frm('Epoch %5d/%5d' % (epoch, epochs))

            # train the model for one epoch
            self.train_epoch(loader=train_loader, loss_fn=loss_fn, optimizer=optimizer, epoch=epoch,
                             augmenter=augmenter, print_stats=print_stats, writer=writer,
                             write_images=epoch % write_images_freq == 0, device=device)

            # adjust learning rate if necessary
            if scheduler is not None:
                scheduler.step(epoch=epoch)

                # and keep track of the learning rate
                writer.add_scalar('learning_rate', float(scheduler.get_lr()[0]), epoch)

            # test the model for one epoch is necessary
            if epoch % test_freq == 0:
                j = self.test_epoch(loader=test_loader, loss_fn=loss_fn, epoch=epoch, writer=writer,
                                    write_images=True, device=device)

                # and save model if higher segmentation performance was obtained
                if j > j_max:
                    j_max = j
                    torch.save(self, os.path.join(log_dir, 'best_checkpoint.pytorch'))

            # save model every epoch
            torch.save(self, os.path.join(log_dir, 'checkpoint.pytorch'))

        writer.close()

    def set_coi(self, coi):
        """
        Adjust the classes of interest of a U-Net, this allows for flexible retraining between classes

        :param coi: tuple indicating the class indices
        """
        # adjust fields
        self.coi = coi
        self.out_channels = len(coi)

        # adjust output channels
        self.decoder.output = nn.Conv3d(self.feature_maps, self.out_channels, kernel_size=1)



from copy import copy

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy.ndimage.morphology import distance_transform_edt
from skimage import measure

from neuralnets.data.datasets import StronglyLabeledStandardDataset, StronglyLabeledVolumeDataset, \
    StronglyLabeledMultiVolumeDataset

from neuralnets.util.tools import tensor_to_device


class CrossEntropyLoss(nn.Module):
    """
    Cross entropy loss function

    :param initalization weight: weights for the classes (C)
    :param forward logits: logits tensor (B, C, N_1, N_2, ...)
    :param forward target: targets tensor (B, N_1, N_2, ...)
    :return: focal loss
    """

    def __init__(self, weight=None, device=0):

        super(CrossEntropyLoss, self).__init__()

        self.weight = weight
        self.device = device

        w = None
        if weight is not None:
            w = tensor_to_device(torch.Tensor(weight), device=device)
        self.ce = nn.CrossEntropyLoss(weight=w, reduction="none")

    def forward(self, logits, target, mask=None):

        # compute loss unreduced and reshape to vector
        loss = self.ce(logits, target).view(-1)

        # size averaging if necessary
        if mask is not None:
            loss = loss[mask.view(-1)].mean()
        else:
            loss = loss.mean()

        return loss

    def set_weight(self, weight=None):

        self.weight = weight

        w = None
        if weight is not None:
            w = tensor_to_device(torch.Tensor(weight), device=self.device)

        self.ce = nn.CrossEntropyLoss(weight=w, reduction="none")


class FocalLoss(nn.Module):
    """
    Focal loss function (T.-Y. Lin, P. Goyal, R. Girshick, K. He, and P. Dollar. Focal Loss for Dense Object Detection, 2017)

    :param initalization alpha: weights for the classes (C)
    :param forward logits: logits tensor (B, C, N_1, N_2, ...)
    :param forward target: targets tensor (B, N_1, N_2, ...)
    :return: focal loss
    """

    def __init__(self, gamma=2, alpha=None):

        super(FocalLoss, self).__init__()

        self.alpha = alpha
        # normalize alpha if necessary
        if self.alpha is not None:
            self.alpha = torch.Tensor(self.alpha / np.sum(self.alpha))
        self.gamma = gamma

    def forward(self, logits, target, mask=None):

        # apply log softmax
        log_p = F.log_softmax(logits, dim=1)
        p = F.softmax(logits, dim=1)

        # channels on the last axis
        input_size = logits.size()
        for d in range(1, len(input_size) - 1):
            log_p = log_p.transpose(d, d + 1)
        log_p = log_p.contiguous()

        # reshape everything
        log_p = log_p[target[:, 0, ...].unsqueeze(-1).repeat_interleave(input_size[1], dim=-1) >= 0]
        log_p = log_p.view(-1, input_size[1])
        p = p.view(-1, input_size[1])
        target = target.view(-1)

        # compute negative log likelihood
        if self.alpha is not None:
            cw = tensor_to_device(self.alpha, device=target.device.index)
        else:
            cw = None
        loss = F.nll_loss((1 - p) ** self.gamma * log_p, target, reduction='none', weight=cw)

        # size averaging if necessary
        if mask is not None:
            loss = loss[mask.view(-1)].mean()
        else:
            loss = loss.mean()

        return loss


class DiceLoss(nn.Module):
    """
    Dice loss function

    :param forward logits: logits tensor (B, C, N_1, N_2, ...)
    :param forward target: targets tensor (B, N_1, N_2, ...)
    :return: dice loss
    """

    def forward(self, logits, target, mask=None):

        # precompute for efficiency
        if mask is not None:
            mask = mask.view(-1)

        # apply softmax and compute dice loss for each class
        probs = F.softmax(logits, dim=1)
        dice = 0
        for c in range(1, logits.size(1)):
            p = probs[:, c:c + 1, ...]
            t = (target == c).long()

            # reshape everything to vectors
            p = p.contiguous().view(-1)
            t = t.contiguous().view(-1)

            # mask if necessary
            if mask is not None:
                p = p[mask]
                t = t[mask]

            # dice loss
            numerator = 2 * torch.sum(p * t)
            denominator = torch.sum(p + t)
            dice = dice + (1 - ((numerator + 1) / (denominator + 1)))

        # compute average
        dice = dice / (logits.size(1) - 1)

        return dice


class TverskyLoss(nn.Module):
    """
    Tversky loss function (S. S. M. Salehi, D. Erdogmus, and A. Gholipour. Tversky loss function for image segmentation using 3D fully convolutional deep networks, 2017)

    :param initialization c: index of the class of index
    :param forward logits: logits tensor (B, C, N_1, N_2, ...)
    :param forward target: targets tensor (B, N_1, N_2, ...)
    :return: tversky loss
    """

    def __init__(self, beta=0.5, c=1):
        super(TverskyLoss, self).__init__()

        self.beta = beta
        self.c = c

    def forward(self, logits, target, mask=None):

        # precompute for efficiency
        if mask is not None:
            mask = mask.view(-1)

        # apply softmax and compute dice loss for each class
        probs = F.softmax(logits, dim=1)
        tversky = 0
        for c in range(1, logits.size(1)):
            p = probs[:, c:c + 1, ...]
            t = (target == c).long()

            # reshape everything to vectors
            p = p.contiguous().view(-1)
            t = t.contiguous().view(-1)

            # mask if necessary
            if mask is not None:
                p = p[mask]
                t = t[mask]

            # dice loss
            numerator = torch.sum(p * t)
            denominator = numerator + self.beta * torch.sum((1 - t) * p) + (1 - self.beta) * torch.sum((1 - p) * t)
            tversky = tversky + (1 - ((numerator + 1) / (denominator + 1)))

        # compute average
        tversky = tversky / (logits.size(1) - 1)

        return tversky


class LpLoss(nn.Module):
    """
    L_p loss function

    :param initalization p: parameter for the loss function
    :param initalization size_average: flag that specifies whether to apply size averaging at the end or not
    :param forward logits: logits tensor (N_1, N_2, ...)
    :param forward target: targets tensor (N_1, N_2, ...)
    :return: L_p loss
    """

    def __init__(self, p=2, size_average=True):
        super(LpLoss, self).__init__()

        self.p = p
        self.size_average = size_average

    def forward(self, input, target):
        target_rec = torch.sigmoid(input)
        loss = torch.pow(torch.sum(torch.pow(torch.abs(target - target_rec), self.p)), 1 / self.p)
        if self.size_average:
            loss = loss / target.numel()
        return loss


class L2Loss(nn.Module):
    """
    L_2 loss function

    :param initalization size_average: flag that specifies whether to apply size averaging at the end or not
    :param forward logits: logits tensor (N_1, N_2, ...)
    :param forward target: targets tensor (N_1, N_2, ...)
    :return: L_2 loss
    """

    def __init__(self, size_average=True):
        super(L2Loss, self).__init__()

        self.size_average = size_average

    def forward(self, input, target):
        target_rec = torch.sigmoid(input)
        loss = torch.sqrt(torch.sum(torch.pow(target - target_rec, 2)))
        if self.size_average:
            loss = loss / target.numel()
        return loss


class MADLoss(nn.Module):
    """
    Mean absolute deviation (MAD) loss function

    :param initalization size_average: flag that specifies whether to apply size averaging at the end or not
    :param forward pred: predictions tensor (N_1, N_2, ...)
    :param forward target: targets tensor (N_1, N_2, ...)
    :return: MAD loss
    """

    def __init__(self, size_average=True):
        super(MADLoss, self).__init__()

        self.size_average = size_average

    def forward(self, pred, target):
        loss = torch.sum(torch.abs(target - pred))
        if self.size_average:
            loss = loss / target.numel()
        return loss


class KLDLoss(nn.Module):
    """
    Kullback Leibler divergence (KLD) loss function

    :param forward mu: mean tensor (N_1, N_2, ...)
    :param forward log: logarithmic variance tensor (N_1, N_2, ...)
    :return: KLD loss
    """

    def forward(self, mu, logvar):

        if mu.data.ndimension() == 4:
            mu = mu.view(mu.size(0), mu.size(1))
        if logvar.data.ndimension() == 4:
            logvar = logvar.view(logvar.size(0), logvar.size(1))

        klds = 1 + logvar - mu.pow(2) - logvar.exp()
        kld = torch.mean(-0.5 * torch.sum(klds, dim=1), dim=0)

        return kld


def boundary_weight_map(labels, sigma=20, w0=1):
    """
    Compute the boundary weight map, according to the the original U-Net paper

    :param labels: input tensor
    :param optional sigma: damping parameter
    :param optional w0: initial value of the weight map
    :return boundary weight map as a tensor
    """

    y = labels.cpu().numpy()
    weight = np.ones(y.shape)
    diag = np.sqrt(weight.shape[2] ** 2 + weight.shape[3] ** 2)

    # for each image in the batch
    for b in range(y.shape[0]):

        # compute connected components
        comps = measure.label(y[b, 0, ...])
        n_comps = np.max(comps)

        if n_comps > 0:  # if there is at least one component
            if n_comps > 1:  # if there are at least two components
                # remove every component separately and compute distance transform
                dtfs = np.zeros((n_comps, comps.shape[0], comps.shape[1]))
                for c in range(n_comps):
                    # compute distance transform
                    y_ = copy(y[b, 0, ...])
                    y_[comps == (1 + c)] = 0  # remove component
                    dtfs[c, ...] = distance_transform_edt(1 - y_)
                dtfs_sorted = np.sort(dtfs, axis=0)
                dtf1 = dtfs_sorted[0, ...]
                dtf2 = dtfs_sorted[1, ...]
            else:
                # compute distance transform
                dtf1 = distance_transform_edt(1 - y[b, 0, ...])
                dtf2 = diag
            # update weight map
            weight[b, 0, ...] = weight[b, 0, ...] + w0 * (1 - y[b, 0, ...]) * np.exp(
                - (dtf1 + dtf2) ** 2 / (2 * sigma ** 2))

    return torch.Tensor(weight).cuda()


def _parse_loss_params(t):
    params = {}
    for s in t:
        key, val = s.split(':')
        if val.count(',') > 0:
            val = val.split(',')
            for i in range(len(val)):
                val[i] = float(val[i])
        else:
            val = float(val)
        params[key] = val

    return params


def get_loss_function(s):
    """
    Returns a loss function according to the settings in a string. The string is formatted as follows:
        s = <loss-function-name>[#<param>:<param-value>#<param>:<param-value>#...]
            parameter values are either
                - scalars
                - vectors (written as scalars, separated by commas)

    :param s: loss function specifier string, formatted as shown on top
    :return: the required loss function
    """
    t = s.lower().replace("-", "_").split('#')
    name = t[0]
    params = _parse_loss_params(t[1:])
    if name == "ce" or name == "cross_entropy":
        return CrossEntropyLoss(**params)
    elif name == "fl" or name == "focal":
        return FocalLoss(**params)
    elif name == "dl" or name == "dice":
        return DiceLoss(**params)
    elif name == "tl" or name == "tversky":
        return TverskyLoss(**params)
    elif name == "l1":
        return nn.L1Loss(**params)
    elif name == "mse":
        return nn.MSELoss(**params)
    elif name == "kld":
        return KLDLoss(**params)


def get_balancing_weights(dataset):
    """
    Returns a set of balancing weights for a specific labeled dataset

    :param dataset: labeled dataset instance of
                        - neuralnets.data.datasets.StronglyLabeledStandardDataset
                        - neuralnets.data.datasets.StronglyLabeledVolumeDataset
                        - neuralnets.data.datasets.StronglyLabeledMultiVolumeDataset
    :return: a tuple of balancing weights, if an unsuitable object is provided, it returns None
    """

    if isinstance(dataset, StronglyLabeledStandardDataset) or isinstance(dataset, StronglyLabeledVolumeDataset):
        weight = np.zeros((len(dataset.coi)))
        for i, c in enumerate(dataset.coi):
            weight[i] = 1 / np.count_nonzero(dataset.labels == c)
        weight = weight / np.sum(weight)
        return tuple(weight)
    elif isinstance(dataset, StronglyLabeledMultiVolumeDataset):
        freq = np.zeros((len(dataset.coi)))
        for i, c in enumerate(dataset.coi):
            for labels in dataset.labels:
                freq[i] += np.count_nonzero(labels == c)
        weight = 1 / freq
        weight = weight / np.sum(weight)
        return weight

    return None



import cv2
import numpy as np
import numpy.random as rnd
import torch
import torch.nn.functional as F
from scipy.ndimage import spline_filter1d, zoom

from neuralnets.util.tools import tensor_to_device


class ToTensor(object):
    """
    Transforms a numpy array into a tensor

    :param initialization device: GPU device where the computations should occur
    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __init__(self, device=0):
        self.device = device

    def __call__(self, x):
        return tensor_to_device(torch.Tensor(x), device=self.device)


class ToFloatTensor(object):
    """
    Transforms a Tensor to a FloatTensor

    :param initialization device: GPU device where the computations should occur
    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __init__(self, device=0):
        self.device = device

    def __call__(self, x):
        return tensor_to_device(x.float(), device=self.device)


class ToLongTensor(object):
    """
    Transforms a Tensor to a LongTensor

    :param initialization device: GPU device where the computations should occur
    :param forward x: input array (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __init__(self, device=0):
        self.device = device

    def __call__(self, x):
        return tensor_to_device(x.long(), device=self.device)


class AddChannelAxis(object):
    """
    Add a channel to the input tensor

    :param forward x: input tensor (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __call__(self, x):
        return x.unsqueeze(0)


class AddNoise(object):
    """
    Adds noise to the input

    :param initialization prob: probability of adding noise
    :param initialization sigma_min: minimum noise standard deviation
    :param initialization sigma_max: maximum noise standard deviation
    :param initialization include_segmentation: 2nd half of the batch will not be augmented as this is assumed to be a (partial) segmentation
    :param forward x: input tensor (B, N_1, N_2, ...)
    :return: output tensor (B, N_1, N_2, ...)
    """

    def __init__(self, prob=0.5, sigma_min=0.0, sigma_max=1.0, include_segmentation=False):
        self.prob = prob
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.include_segmentation = include_segmentation

    def __call__(self, x):

        if rnd.rand() < self.prob:
            sigma = rnd.uniform(self.sigma_min, self.sigma_max)
            if self.include_segmentation:
                sz = np.asarray(x.size())
                sz[0] = sz[0] // 2
                sz = tuple(sz)
                noise = torch.cat((torch.normal(0, sigma, sz), torch.zeros(sz)), dim=0)
            else:
                noise = torch.normal(0, sigma, x.size())
            noise = tensor_to_device(noise, device=x.device.index)
            return x + noise
        else:
            return x


class Normalize(object):
    """
    Normalizes the input

    :param type: the desired type of normalization (z, unit or minmax)
    :param optional bits: number of bits used to represent a pixel value (only if type is unit)
    :param optional mu: normalization mean (only if type is z)
    :param optional sigma: normalization std (only if type is z)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __init__(self, type='unit', bits=8, mu=None, sigma=None):
        self.type = type
        self.bits = bits
        self.mu = mu
        self.sigma = sigma

    def __call__(self, x):
        if self.type == 'minmax':
            # apply minmax normalization
            m = x.min()
            M = x.max()
            eps = 1e-5
            return (x - m + eps) / (M - m + eps)
        elif self.type == 'unit':
            # apply unit normalization
            return x / (2**self.bits)
        else:
            # apply z normalization
            mu = torch.mean(x) if self.mu is None else self.mu
            sigma = torch.std(x) if self.sigma is None else self.sigma
            return (x - mu) / sigma


class ContrastAdjust(object):
    """
    Apply contrast adjustments to the data

    :param initialization prob: probability of adjusting contrast
    :param initialization adj: maximum adjustment (maximum intensity shift for minimum and maximum the new histogram)
    :param initialization include_segmentation: 2nd half of the batch will not be augmented as this is assumed to be a (partial) segmentation
    :param forward x: input tensor (N_1, N_2, N_3, ...)
    :return: output tensor (N_1, N_2, N_3, ...)
    """

    def __init__(self, prob=1, adj=0.2, include_segmentation=False):
        self.prob = prob
        self.adj = adj
        self.include_segmentation = include_segmentation

    def __call__(self, x):

        if rnd.rand() < self.prob:
            x_ = x
            if self.include_segmentation:
                x_ = x[:x.size(0) // 2, ...]

            m = x_.min()
            M = x_.max()
            m_ = 2 * self.adj * torch.rand(1) - self.adj + m
            M_ = 2 * self.adj * torch.rand(1) - self.adj + M
            m_ = m_.cuda(device=x.device)
            M_ = M_.cuda(device=x.device)
            x_adj = ((x[:x.size(0) // 2, ...] - m) / (M - m)) * (M_ - m_) + m_

            if self.include_segmentation:
                return torch.cat((x_adj, x[x.size(0) // 2:, ...]), dim=0)
            else:
                return ((x - m) / (M - m)) * (M_ - m_) + m_
        else:
            return x


class Scale(object):
    """
    Scales the input by a specific factor (randomly selected from a minimum-maximum range)

    :param initialization scale_factor: minimum and maximum scaling factor
    :param forward x: input tensor (B, C, [Z  , Y  ,] X)
    :return: output tensor (B, C, [Z' , Y' ,] X)
    """

    def __init__(self, scale_factor=(0.5, 1.5), mode='bilinear'):
        self.scale_factor = scale_factor
        self.mode = mode

    def __call__(self, x):
        if type(self.scale_factor) == tuple:
            scale_factor = (self.scale_factor[1] - self.scale_factor[0]) * np.random.random_sample() + \
                           self.scale_factor[0]
        else:
            scale_factor = self.scale_factor
        return F.interpolate(x, scale_factor=scale_factor, mode=self.mode, align_corners=False)


class FlipX(object):
    """
    Perform a flip along the X axis

    :param initialization prob: probability of flipping
    :param forward x: input tensor (B, C, N_1, N_2, ...)
    :return: output tensor (B, C, N_1, N_2, ...)
    """

    def __init__(self, prob=1):
        self.prob = prob

    def __call__(self, x):

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 1])
        else:
            return x


class FlipY(object):
    """
    Perform a flip along the Y axis

    :param initialization prob: probability of flipping
    :param forward x: input tensor (B, C, N_1, N_2, ...)
    :return: output tensor (B, C, N_1, N_2, ...)
    """

    def __init__(self, prob=1):
        self.prob = prob

    def __call__(self, x):

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 2])
        else:
            return x


class FlipZ(object):
    """
    Perform a flip along the Z axis

    :param initialization prob: probability of flipping
    :param forward x: input tensor (B, C, N_1, N_2, ...)
    :return: output tensor (B, C, N_1, N_2, ...)
    """

    def __init__(self, prob=1):
        self.prob = prob

    def __call__(self, x):

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 3])
        else:
            return x


class Rotate90(object):
    """
    Rotate the inputs by 90 degree angles

    :param initialization prob: probability of rotating
    :param forward x: input tensor (B, C, N_1, N_2, ...)
    :return: output tensor (B, C, N_1, N_2, ...)
    """

    def __init__(self, prob=1):
        self.prob = prob

    def __call__(self, x):

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.rot90(x, k=rnd.randint(0, 4), dims=[n - 2, n - 1])
        else:
            return x


class RotateRandom_2D(object):
    """
    Rotate the inputs by a random amount of degrees within interval.

    :param initialization shape: 2D shape of the input image
    :param initialization rng: random degree interval size (symmetric around 0)
    :param initialization device: GPU device where the computations should occur
    :param forward x: input tensor (B, C, Y , X)
    :return: output tensor (B, C, Y , X)
    """

    def __init__(self, shape, prob=1.0, rng=200, device=0):
        self.shape = tuple(shape)
        self.device = device
        self.rng = int(rng / 2)
        self.prob = prob
        self.image_center = int(self.shape[0] / 2), int(self.shape[1] / 2)

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        self.xv, self.yv = np.meshgrid(i, j)

    def _rotation_grid(self):
        angle = np.random.randint(-self.rng, self.rng)
        rot_matrix = cv2.getRotationMatrix2D(self.image_center, angle, 1.0)
        xv = cv2.warpAffine(self.xv, rot_matrix, self.xv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        yv = cv2.warpAffine(self.yv, rot_matrix, self.yv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)

        grid = torch.cat((torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1)), dim=-1)
        grid = tensor_to_device(grid.unsqueeze(0), device=self.device)
        return grid

    def __call__(self, x):

        if rnd.rand() < self.prob:
            grid = self._rotation_grid()
            grid = grid.repeat_interleave(x.size(0), dim=0)
            return F.grid_sample(x, grid)
        else:
            return x


class RotateRandom_3D(object):
    """
    Rotate the inputs by a random amount of degrees within interval.

    :param initialization shape: 3D shape of the input image
    :param initialization rng: random degree interval size (symmetric around 0)
    :param initialization device: GPU device where the computations should occur
    :param forward x: input tensor (B, C, Z, Y , X)
    :return: output tensor (B, C, Z, Y , X)
    """

    def __init__(self, shape, prob=1.0, rng=200, device=0):
        self.shape = tuple(shape)
        self.device = device
        self.rng = int(rng / 2)
        self.prob = prob
        self.image_center = int(self.shape[0] / 2), int(self.shape[1] / 2), int(self.shape[2] / 2)

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        k = np.linspace(-1, 1, shape[2])
        self.xv, self.yv, self.zv = np.meshgrid(i, j, k)

    def _rotation_grid(self):
        angle = np.random.randint(-self.rng, self.rng)
        rot_matrix = cv2.getRotationMatrix2D(self.image_center, angle, 1.0)
        xv = cv2.warpAffine(self.xv, rot_matrix, self.xv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        yv = cv2.warpAffine(self.yv, rot_matrix, self.yv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        zv = cv2.warpAffine(self.zv, rot_matrix, self.zv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)

        grid = torch.cat(
            (torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1), torch.Tensor(zv).unsqueeze(-1)), dim=-1)
        grid = tensor_to_device(grid.unsqueeze(0), device=self.device)
        return grid

    def __call__(self, x):

        if rnd.rand() < self.prob:
            grid = self._rotation_grid()
            grid = grid.repeat_interleave(x.size(0), dim=0)
            return F.grid_sample(x, grid)
        else:
            return x


class RandomCrop_2D(object):
    """
    Selects a random crop from the input

    :param initialization crop_shape: 2D shape of the crop
    :param forward x: input tensor (B, C, Y , X)
    :return: output tensor (B, C, Y', X')
    """

    def __init__(self, crop_shape):
        self.crop_shape = crop_shape

    def __call__(self, x):
        r = np.random.randint(0, x.size(2) - self.crop_shape[0] + 1)
        c = np.random.randint(0, x.size(3) - self.crop_shape[1] + 1)
        return x[:, :, r:r + self.crop_shape[0], c:c + self.crop_shape[1]]


class RandomCrop_3D(object):
    """
    Selects a random crop from the input

    :param initialization crop_shape: 3D shape of the crop
    :param forward x: input tensor (B, C, Z, Y , X)
    :return: output tensor (B, C, Z', Y', X')
    """

    def __init__(self, crop_shape):
        self.crop_shape = crop_shape

    def __call__(self, x):
        r = np.random.randint(0, x.size(2) - self.crop_shape[0] + 1)
        c = np.random.randint(0, x.size(3) - self.crop_shape[1] + 1)
        z = np.random.randint(0, x.size(4) - self.crop_shape[2] + 1)
        return x[:, :, r:r + self.crop_shape[0], c:c + self.crop_shape[1], z:z + self.crop_shape[2]]


class RandomDeformation_2D(object):
    """
    Apply random deformation to the inputs

    :param initialization shape: shape of the inputs
    :param initialization prob: probability of deforming the data
    :param initialization device: GPU device where the computations should occur
    :param initialization points: seed points for deformation
    :param initialization grid_size: tuple with number of pixels between each grid point
    :param initialization n_grids: number of grids to load in advance (chose more for higher variance in the data)
    :param initialization include_segmentation: 2nd half of the batch needs casting to integers because of warping
    :param forward x: input tensor (B, C, Y , X)
    :return: output tensor (B, C, Y, X)
    """

    def __init__(self, shape, prob=1, device=0, points=None, grid_size=(64, 64), sigma=0.01, n_grids=1000,
                 include_segmentation=False):
        self.shape = shape
        self.prob = prob
        self.device = device
        self.grid_size = grid_size
        if points == None:
            points = [shape[0] // self.grid_size[0], shape[1] // self.grid_size[1]]
        self.points = points
        self.sigma = sigma
        self.n_grids = n_grids
        self.grids = []
        self.include_segmentation = include_segmentation

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        xv, yv = np.meshgrid(i, j)

        grid = torch.cat((torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1)), dim=-1)
        grid = tensor_to_device(grid.unsqueeze(0), device=self.device)
        self.grid = grid

        # generate several random grids in advance (less CPU work)
        for i in range(self.n_grids):
            self.grids.append(self._deformation_grid())

    def _deformation_grid(self):
        sigma = np.random.rand() * self.sigma
        displacement = np.random.randn(*self.points, 2) * sigma

        # filter the displacement
        displacement_f = np.zeros_like(displacement)
        for d in range(0, displacement.ndim - 1):
            spline_filter1d(displacement, axis=d, order=3, output=displacement_f, mode='nearest')
            displacement = displacement_f

        # resample to proper size
        displacement_f = np.zeros((self.shape[0], self.shape[1], 2))
        for d in range(0, displacement.ndim - 1):
            displacement_f[:, :, d] = zoom(displacement[:, :, d], self.grid_size)

        displacement = tensor_to_device(torch.Tensor(displacement_f).unsqueeze(0), device=self.device)
        grid = self.grid + displacement

        return grid

    def __call__(self, x):

        if rnd.rand() < self.prob:
            grid = self.grids[rnd.randint(self.n_grids)]
            if x.size()[2:4] != grid.size()[1:3]:  # reshape grid if necessary
                grid = F.interpolate(grid.permute(0, 3, 1, 2), x.size()[2:4]).permute(0, 2, 3, 1)
            grid = grid.repeat_interleave(x.size(0), dim=0)
            x_aug = F.grid_sample(x, grid, padding_mode="border")
            if self.include_segmentation:
                x_aug[x.size(0) // 2:, ...] = torch.round(x_aug[x.size(0) // 2:, ...])
            return x_aug
        else:
            return x


class RandomDeformation_3D(object):
    """
    Apply random deformation to the inputs

    :param initialization shape: shape of the inputs
    :param initialization prob: probability of deforming the data
    :param initialization device: GPU device where the computations should occur
    :param initialization points: seed points for deformation
    :param initialization grid_size: tuple with number of pixels between each grid point
    :param initialization n_grids: number of grids to load in advance (chose more for higher variance in the data)
    :param initialization include_segmentation: 2nd half of the batch needs casting to integers because of warping
    :param forward x: input tensor (B, C, Z, Y , X)
    :return: output tensor (B, C, Z, Y, X)
    """

    def __init__(self, shape, prob=1, device=0, points=None, grid_size=(64, 64), sigma=0.01, n_grids=1000,
                 include_segmentation=False):
        self.shape = shape
        self.prob = prob
        self.device = device
        self.grid_size = grid_size
        if points == None:
            points = [shape[0] // self.grid_size[0], shape[1] // self.grid_size[1]]
        self.points = points
        self.sigma = sigma
        self.n_grids = n_grids
        self.grids = []
        self.include_segmentation = include_segmentation

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        xv, yv = np.meshgrid(i, j)

        grid = torch.cat((torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1)), dim=-1)
        grid = tensor_to_device(grid.unsqueeze(0), device=self.device)
        self.grid = grid

        # generate several random grids in advance (less CPU work)
        for i in range(self.n_grids):
            self.grids.append(self._deformation_grid())

    def _deformation_grid(self):
        sigma = np.random.rand() * self.sigma
        displacement = np.random.randn(*self.points, 2) * sigma

        # filter the displacement
        displacement_f = np.zeros_like(displacement)
        for d in range(0, displacement.ndim - 1):
            spline_filter1d(displacement, axis=d, order=3, output=displacement_f, mode='nearest')
            displacement = displacement_f

        # resample to proper size
        displacement_f = np.zeros((self.shape[0], self.shape[1], 2))
        for d in range(0, displacement.ndim - 1):
            displacement_f[:, :, d] = zoom(displacement[:, :, d], self.grid_size)

        displacement = tensor_to_device(torch.Tensor(displacement_f).unsqueeze(0), device=self.device)
        grid = self.grid + displacement

        return grid

    def __call__(self, x):

        if rnd.rand() < self.prob:
            grid = self.grids[rnd.randint(self.n_grids)]
            if x.size()[3:5] != grid.size()[1:3]:  # reshape grid if necessary
                grid = F.interpolate(grid.permute(0, 3, 1, 2), x.size()[3:5]).permute(0, 2, 3, 1)
            grid = grid.repeat_interleave(x.size(0) * x.size(2), dim=0)
            x_aug = F.grid_sample(torch.reshape(x, (x.size(0) * x.size(2), x.size(1), x.size(3), x.size(4))), grid,
                                  padding_mode="border")
            x_aug = torch.reshape(x_aug, x.size())
            if self.include_segmentation:
                x_aug[x.size(0) // 2:, ...] = torch.round(x_aug[x.size(0) // 2:, ...])
            return x_aug
        else:
            return x



import numpy as np
from scipy.spatial.distance import directed_hausdorff


def jaccard(y_true, y_pred, w=None):
    """
    Jaccard index between two segmentations

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the Jaccard index
    """

    # check mask
    if w is None:
        w = np.ones_like(y_true, dtype='bool')
    y_true = y_true[w]
    y_pred = y_pred[w]

    # binarize
    y_true = (y_true > 0.5).astype('int')
    y_pred = (y_pred > 0.5).astype('int')

    # compute jaccard score
    intersection = np.sum(y_true * y_pred)
    union = np.sum(y_true) + np.sum(y_pred) - intersection

    return intersection / union


def dice(y_true, y_pred, w=None):
    """
    Dice coefficient between two segmentations

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the Jaccard index
    """

    j = jaccard(y_true, y_pred, w=w)

    return 2 * j / (1 + j)


def accuracy_metrics(y_true, y_pred, w=None):
    """
    Accuracy metrics between two segmentations (accuracy, balanced accuracy, precision, recall and f1-score)

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the Jaccard index
    """

    # check mask
    if w is None:
        w = np.ones_like(y_true, dtype='bool')
    y_true = y_true[w]
    y_pred = y_pred[w]

    # binarize
    y_true = (y_true > 0.5).astype('int')
    y_pred = (y_pred > 0.5).astype('int')

    # compute accuracy metrics
    tp = np.sum(y_true * y_pred)
    tn = np.sum((1 - y_true) * (1 - y_pred))
    fp = np.sum((1 - y_true) * y_pred)
    fn = np.sum(y_true * (1 - y_pred))
    total = tp + tn + fp + fn
    accuracy = (tp + tn) / total
    recall = tp / (tp + fn)
    specificity = tn / (tn + fp)
    balanced_accuracy = (recall + specificity) / 2
    precision = tp / (tp + fp)
    f1 = 2 * (precision * recall) / (precision + recall)

    return accuracy, balanced_accuracy, precision, recall, f1


def hausdorff_distance(x, y):
    """
    Hausdorff distance between two segmentations

    :param x: array
    :param y: array
    :return: the hausdorff distance
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    hd_0 = 0
    hd_1 = 0
    hd = 0
    for i in range(x.shape[0]):
        hd_0 += directed_hausdorff(x[i, ...], y[i, ...])[0]
        hd_1 += directed_hausdorff(y[i, ...], x[i, ...])[0]
        hd += max(hd_0, hd_1)
    hd_0 /= x.shape[0]
    hd_1 /= x.shape[0]
    hd /= x.shape[0]

    return hd, hd_0, hd_1



# This script illustrates how to analyze a labeled dataset
import os

import numpy as np

from neuralnets.util.io import read_volume, write_volume

# path to the data and the corresponding type
data_path = '/home/jorisro/research/data/EM/VIB/Project_093_WimAnnaert/2017_04_07_Po93_6Q_46'
type = 'pngseq'

# analyze the data
for ds in ['registered']:
    # read the data
    data = read_volume(os.path.join(data_path, ds), type=type)
    # measure size
    sz = data.shape
    # report
    print("Dataset %s: " % (ds))
    print("    Size: %d x %d x %d" % (sz[0], sz[1], sz[2]))
